/*	$Id: krecaptcha.c,v 1.1.1.1 2014/05/08 10:21:15 kristaps Exp $ */
/*
 * Copyright (c) 2014 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <curl/curl.h>

#include "krecaptcha.h"

/*
 * Buffer maintained by the function reading results from the server
 * following our POST.
 */
struct	buf {
	char		*buf; /* always nil-terminated */
	size_t		 sz; /* length not including nil */
};

/*
 * Read response data from the pipe.
 * This must be very careful not to overrun any data.
 */
static size_t 
dobuf(void *ptr, size_t sz, size_t nm, void *arg)
{
	struct buf	*buf = arg;
	size_t		 rsz, i;
	char		*cp;
	void		*copy;

	/* Obviously, we don't care about zero-length. */
	if (0 == sz || 0 == nm)
		return(0);

	/* Check for multiplication overflow. */
	if (SIZE_MAX / nm < sz) 
		return(0);
	rsz = sz * nm;

	/* Now check for additive overflow. */
	if (SIZE_MAX - rsz < buf->sz)
		return(0);

	/* 
	 * Actually reallocate our buffer.
	 * Make sure that we leave room for nil-termination.
	 */
	copy = realloc(buf->buf, buf->sz + rsz + 1);
	if (NULL == copy)
		return(0);
	buf->buf = copy;

	/* Copy the data over.  Don't let nil slip by. */
	for (cp = ptr, i = 0; i < rsz; i++) {
		if ('\0' == cp[i])
			return(0);
		buf->buf[buf->sz++] = cp[i];
	}

	/* Nil-terminate. */
	buf->buf[buf->sz] = '\0';
	return(rsz);
}

/*
 * See krecaptcha(3).
 */
enum krecaptchae
krecaptcha(const char *privkey, 
	const char *remote, 
	const char *challenge, 
	const char *response)
{
	CURL		     *curl;
	CURLcode	      res;
	struct curl_httppost *formpost;
	struct curl_httppost *lastptr;
	const char	     *surl;
	struct buf	      buf;
	char		     *cp, *cp2;

	/*
	 * All inputs shall be non-NULL.
	 */
	assert(NULL != remote);
	assert(NULL != challenge);
	assert(NULL != response);
	assert(NULL != privkey);

	formpost = lastptr = NULL;
	surl = "https://api-verify.recaptcha.net/verify";
	memset(&buf, 0, sizeof(struct buf));

	if (NULL == (curl = curl_easy_init()))
		return(KRECAPTCHA_CURL_INIT);

	/*
	 * Add each necessary element to the curl form.
	 * Make sure that we free all our data if we error out.
	 */
	if (CURL_FORMADD_OK != curl_formadd
			(&formpost, &lastptr,
			CURLFORM_COPYNAME, "privatekey",
			CURLFORM_COPYCONTENTS, privkey,
			CURLFORM_END)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	} else if (CURL_FORMADD_OK != curl_formadd
			(&formpost, &lastptr,
			CURLFORM_COPYNAME, "remoteip",
			CURLFORM_COPYCONTENTS, remote,
			CURLFORM_END)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	} else if (CURL_FORMADD_OK != curl_formadd
			(&formpost, &lastptr,
			CURLFORM_COPYNAME, "challenge",
			CURLFORM_COPYCONTENTS, challenge,
			CURLFORM_END)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	} else if (CURL_FORMADD_OK != curl_formadd
			(&formpost, &lastptr,
			CURLFORM_COPYNAME, "response",
			CURLFORM_COPYCONTENTS, response,
			CURLFORM_END)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	} else if (CURLE_OK != curl_easy_setopt
			(curl, CURLOPT_URL, surl)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	} else if (CURLE_OK != curl_easy_setopt
			(curl, CURLOPT_HTTPPOST, formpost)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	} else if (CURLE_OK != curl_easy_setopt
			(curl, CURLOPT_SSL_VERIFYPEER, 0L)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	} else if (CURLE_OK != curl_easy_setopt
			(curl, CURLOPT_WRITEFUNCTION, dobuf)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	} else if (CURLE_OK != curl_easy_setopt
			(curl, CURLOPT_WRITEDATA, &buf)) {
		curl_formfree(formpost);
		curl_easy_cleanup(curl);
		return(KRECAPTCHA_CURL_SETUP);
	}

	/* 
	 * Actually run the query.
	 * Then clean up everything associated with the query.
	 */
	res = curl_easy_perform(curl);
	curl_formfree(formpost);
	curl_easy_cleanup(curl);
	curl_global_cleanup();

	/* Weren't able to process... (many possible reasons) */
	if (CURLE_OK != res) { 
		free(buf.buf);
		return(KRECAPTCHA_CURL_PERFORM);
	}

	/*
	 * Check for line one.
	 * This consists of either "true" or "false".
	 */
	if (NULL == (cp = strchr(buf.buf, '\n'))) {
		free(buf.buf);
		return(KRECAPTCHA_RESPONSE);
	} else 
		*cp++ = '\0';

	if (0 == strcmp(buf.buf, "true")) {
		free(buf.buf);
		return(KRECAPTCHA_OK);
	} else if (strcmp(buf.buf, "false")) {
		free(buf.buf);
		return(KRECAPTCHA_RESPONSE);
	} 

	/*
	 * Now that we have an error, check line 2.
	 * This consists of an error code.
	 */
	if (NULL != (cp2 = strchr(cp, '\n')))
		*cp2 = '\0';

	if (0 == strcmp(cp, "invalid-site-private-key")) {
		free(buf.buf);
		return(KRECAPTCHA_PRIVKEY);
	} else if (0 == strcmp(cp, "invalid-request-cookie")) {
		free(buf.buf);
		return(KRECAPTCHA_PRIVKEY);
	} else if (0 == strcmp(cp, "incorrect-captcha-sol")) {
		free(buf.buf);
		return(KRECAPTCHA_PRIVKEY);
	} else if (0 == strcmp(cp, "captcha-timeout")) {
		free(buf.buf);
		return(KRECAPTCHA_PRIVKEY);
	}

	free(buf.buf);
	return(KRECAPTCHA_RESPONSE);
}
